
#include "spdcc.h"

#define SPDC_RANK 4

/*----------------------------------------------------------------------------
 * Function declarations
 *--------------------------------------------------------------------------*/

int spdc_prepare_working_arrays(struct spdc_mdarray *arrays[12],
				struct spdc_prm prm);
int spdc_calculate_field_parameters(struct field_co *flds, struct spdc_prm prm);
int spdc_dump_all_arrays(spdc_mdarray ** arrays, int n_arrays);
int spdc_prepare_input_fields(spdc_mdarray * rho, spdc_mdarray * u0,
			      struct spdc_prm prm);
int spdc_compute_q_space(spdc_mdarray * q, struct spdc_prm prm);
int spdc_fftw(spdc_mdarray * arr, int direction, double complex normalization);

int spdc_phi(spdc_mdarray * phi, spdc_mdarray * v, spdc_mdarray * q, struct
	     field_co *flds, struct spdc_prm prm);
int spdc_near_field_offset(spdc_mdarray * A, spdc_mdarray * q, field_co * flds,
			   struct spdc_prm prm);
extern inline double complex spdc_compute_delta_kz_type2(spdc_mdarray * q,
							 ptrdiff_t i,
							 ptrdiff_t j,
							 ptrdiff_t k,
							 ptrdiff_t l,
							 struct field_co
							 *flds);
extern inline double complex spdc_compute_delta_kz_type1(spdc_mdarray * q,
							 ptrdiff_t i,
							 ptrdiff_t j,
							 ptrdiff_t k,
							 ptrdiff_t l,
							 struct field_co
							 *flds);
extern inline double complex spdc_compute_delta_kz_type0(spdc_mdarray * q,
							 ptrdiff_t i,
							 ptrdiff_t j,
							 ptrdiff_t k,
							 ptrdiff_t l,
							 struct field_co
							 *flds);
extern inline double complex spdc_pump_spectrum(spdc_mdarray * v,
						struct spdc_prm prm,
						ptrdiff_t i, ptrdiff_t j,
						ptrdiff_t k, ptrdiff_t l);
int spdc_schmidt_decomposition(spdc_mdarray * phi, size_t k,
			       struct spdc_prm prm);

/*----------------------------------------------------------------------------
 * Command line argument parsing
 *--------------------------------------------------------------------------*/

const char *argp_program_version = VERSION;
const char *argp_program_bug_address =
    "Robert Elsner <robert.elsner@uni-potsdam.de>";
static char doc[] =
    "Calculation of transverse spatial effects of SPDC biphotons";
static char args_doc[] = "";

/* define options */
/*[[[cog
import cog_codegen

options = cog_codegen.get_options()
cog_codegen.generate_argp_code(options)
]]] */
static struct argp_option options[] = {
	{"lambda-pump", 4000, "FLOAT", 0, "pump beam vacuum wavelength"},
	{"lambda-signal", 4001, "FLOAT", 0, "signal beam vacuum wavelength"},
	{"lambda-idler", 4002, "FLOAT", 0, "idler beam vacuum wavelength"},
	{"hermite-gauss-m", 4003, "INT", 0,
	 "m order of the hermite-gauss TEM_mn pump mode"},
	{"hermite-gauss-n", 4004, "INT", 0,
	 "n order of the hermite-gauss TEM_mn pump mode"},
	{"pump-w0", 4005, "FLOAT", 0, "w0 of the pump beam"},
	{"pump-z", 4006, "FLOAT", 0, "z of the pump beam"},
	{"pseudo-tem", 4007, "INT", 0,
	 "Use a pseuo-TEM mode generated from an TEM via a mode converter"},
	{"width", 4008, "FLOAT", 0, "Transverse crystal width"},
	{"length", 4009, "FLOAT", 0, "Crystal length in z-direction"},
	{"theta", 4010, "FLOAT", 0,
	 "Angle of the crystal's OA wrt to the z-axis in degree"},
	{"phase", 4011, "INT", 0, "Enable exponential phase matching term"},
	{"phase-compensation", 4012, "INT", 0, "Enable exp phase compensation"},
	{"phase-match-type", 4013, "INT", 0,
	 "Set phase II (default) or I phase matching"},
	{"N", 4014, "INT", 0, "Number of discrete samples in each dimension"},
	{"schmidt-rank", 4015, "INT", 0,
	 "Rank of the approximate Schmidt decomposition"},
	{"eigenvalues-only", 4016, "INT", 0, "Only output the eigenvalues S"},
	{"focal-length", 4017, "FLOAT", 0,
	 "Focal length of the lenses of the 4F correlator"},
	{"nf-signal-offset", 4018, "FLOAT", 0,
	 "Longitudinal near field signal offset"},
	{"nf-idler-offset", 4019, "FLOAT", 0,
	 "Longitudinal near field idler offset"},
	{"ff-signal-offset", 4020, "FLOAT", 0,
	 "Longitudinal far field signal offset"},
	{"ff-idler-offset", 4021, "FLOAT", 0,
	 "Longitudinal far field idler offset"},
	{"knife-edge-pos-pump-px", 4022, "FLOAT", 0,
	 "A knife edge aperture at pump beam plane pump-pump in x-direction"},
	{"knife-edge-pos-pump-py", 4023, "FLOAT", 0,
	 "A knife edge aperture at pump beam plane pump-pump in y-direction"},
	{"knife-edge-invert-pump-px", 4024, "INT", 0,
	 "A knife edge aperture at pump beam plane pump-pump in x-direction"},
	{"knife-edge-invert-pump-py", 4025, "INT", 0,
	 "A knife edge aperture at pump beam plane pump-pump in y-direction"},
	{"circ-ap-r-pump-p", 4026, "FLOAT", 0,
	 "circular aperture mask at pump beam plane pump-pump"},
	{"circ-ap-cntr-x-pump-p", 4027, "FLOAT", 0,
	 "circular aperture mask at pump beam plane pump-pump"},
	{"circ-ap-cntr-y-pump-p", 4028, "FLOAT", 0,
	 "circular aperture mask at pump beam plane pump-pump"},
	{"single-slit-cntr-pump-px", 4029, "FLOAT", 0,
	 "a single-slit aperture at pump beam plane pump-pump in x-direction"},
	{"single-slit-cntr-pump-py", 4030, "FLOAT", 0,
	 "a single-slit aperture at pump beam plane pump-pump in y-direction"},
	{"single-slit-width-pump-px", 4031, "FLOAT", 0,
	 "a single-slit aperture at pump beam plane pump-pump in x-direction"},
	{"single-slit-width-pump-py", 4032, "FLOAT", 0,
	 "a single-slit aperture at pump beam plane pump-pump in y-direction"},
	{"double-slit-sep-pump-px", 4033, "FLOAT", 0,
	 "a double-slit aperture at pump beam plane pump-pump in x-direction"},
	{"double-slit-sep-pump-py", 4034, "FLOAT", 0,
	 "a double-slit aperture at pump beam plane pump-pump in y-direction"},
	{"double-slit-width-pump-px", 4035, "FLOAT", 0,
	 "a double-slit aperture at pump beam plane pump-pump in x-direction"},
	{"double-slit-width-pump-py", 4036, "FLOAT", 0,
	 "a double-slit aperture at pump beam plane pump-pump in y-direction"},
	{"knife-edge-pos-f1-sx", 4037, "FLOAT", 0,
	 "A knife edge aperture at far-field SPDC plane f1-signal in x-direction"},
	{"knife-edge-pos-f1-sy", 4038, "FLOAT", 0,
	 "A knife edge aperture at far-field SPDC plane f1-signal in y-direction"},
	{"knife-edge-invert-f1-sx", 4039, "INT", 0,
	 "A knife edge aperture at far-field SPDC plane f1-signal in x-direction"},
	{"knife-edge-invert-f1-sy", 4040, "INT", 0,
	 "A knife edge aperture at far-field SPDC plane f1-signal in y-direction"},
	{"circ-ap-r-f1-s", 4041, "FLOAT", 0,
	 "circular aperture mask at far-field SPDC plane f1-signal"},
	{"circ-ap-cntr-x-f1-s", 4042, "FLOAT", 0,
	 "circular aperture mask at far-field SPDC plane f1-signal"},
	{"circ-ap-cntr-y-f1-s", 4043, "FLOAT", 0,
	 "circular aperture mask at far-field SPDC plane f1-signal"},
	{"single-slit-cntr-f1-sx", 4044, "FLOAT", 0,
	 "a single-slit aperture at far-field SPDC plane f1-signal in x-direction"},
	{"single-slit-cntr-f1-sy", 4045, "FLOAT", 0,
	 "a single-slit aperture at far-field SPDC plane f1-signal in y-direction"},
	{"single-slit-width-f1-sx", 4046, "FLOAT", 0,
	 "a single-slit aperture at far-field SPDC plane f1-signal in x-direction"},
	{"single-slit-width-f1-sy", 4047, "FLOAT", 0,
	 "a single-slit aperture at far-field SPDC plane f1-signal in y-direction"},
	{"double-slit-sep-f1-sx", 4048, "FLOAT", 0,
	 "a double-slit aperture at far-field SPDC plane f1-signal in x-direction"},
	{"double-slit-sep-f1-sy", 4049, "FLOAT", 0,
	 "a double-slit aperture at far-field SPDC plane f1-signal in y-direction"},
	{"double-slit-width-f1-sx", 4050, "FLOAT", 0,
	 "a double-slit aperture at far-field SPDC plane f1-signal in x-direction"},
	{"double-slit-width-f1-sy", 4051, "FLOAT", 0,
	 "a double-slit aperture at far-field SPDC plane f1-signal in y-direction"},
	{"knife-edge-pos-f1-ix", 4052, "FLOAT", 0,
	 "A knife edge aperture at far-field SPDC plane f1-idler in x-direction"},
	{"knife-edge-pos-f1-iy", 4053, "FLOAT", 0,
	 "A knife edge aperture at far-field SPDC plane f1-idler in y-direction"},
	{"knife-edge-invert-f1-ix", 4054, "INT", 0,
	 "A knife edge aperture at far-field SPDC plane f1-idler in x-direction"},
	{"knife-edge-invert-f1-iy", 4055, "INT", 0,
	 "A knife edge aperture at far-field SPDC plane f1-idler in y-direction"},
	{"circ-ap-r-f1-i", 4056, "FLOAT", 0,
	 "circular aperture mask at far-field SPDC plane f1-idler"},
	{"circ-ap-cntr-x-f1-i", 4057, "FLOAT", 0,
	 "circular aperture mask at far-field SPDC plane f1-idler"},
	{"circ-ap-cntr-y-f1-i", 4058, "FLOAT", 0,
	 "circular aperture mask at far-field SPDC plane f1-idler"},
	{"single-slit-cntr-f1-ix", 4059, "FLOAT", 0,
	 "a single-slit aperture at far-field SPDC plane f1-idler in x-direction"},
	{"single-slit-cntr-f1-iy", 4060, "FLOAT", 0,
	 "a single-slit aperture at far-field SPDC plane f1-idler in y-direction"},
	{"single-slit-width-f1-ix", 4061, "FLOAT", 0,
	 "a single-slit aperture at far-field SPDC plane f1-idler in x-direction"},
	{"single-slit-width-f1-iy", 4062, "FLOAT", 0,
	 "a single-slit aperture at far-field SPDC plane f1-idler in y-direction"},
	{"double-slit-sep-f1-ix", 4063, "FLOAT", 0,
	 "a double-slit aperture at far-field SPDC plane f1-idler in x-direction"},
	{"double-slit-sep-f1-iy", 4064, "FLOAT", 0,
	 "a double-slit aperture at far-field SPDC plane f1-idler in y-direction"},
	{"double-slit-width-f1-ix", 4065, "FLOAT", 0,
	 "a double-slit aperture at far-field SPDC plane f1-idler in x-direction"},
	{"double-slit-width-f1-iy", 4066, "FLOAT", 0,
	 "a double-slit aperture at far-field SPDC plane f1-idler in y-direction"},
	{"knife-edge-pos-f2-sx", 4067, "FLOAT", 0,
	 "A knife edge aperture at near-field SPDC plane f2-signal in x-direction"},
	{"knife-edge-pos-f2-sy", 4068, "FLOAT", 0,
	 "A knife edge aperture at near-field SPDC plane f2-signal in y-direction"},
	{"knife-edge-invert-f2-sx", 4069, "INT", 0,
	 "A knife edge aperture at near-field SPDC plane f2-signal in x-direction"},
	{"knife-edge-invert-f2-sy", 4070, "INT", 0,
	 "A knife edge aperture at near-field SPDC plane f2-signal in y-direction"},
	{"circ-ap-r-f2-s", 4071, "FLOAT", 0,
	 "circular aperture mask at near-field SPDC plane f2-signal"},
	{"circ-ap-cntr-x-f2-s", 4072, "FLOAT", 0,
	 "circular aperture mask at near-field SPDC plane f2-signal"},
	{"circ-ap-cntr-y-f2-s", 4073, "FLOAT", 0,
	 "circular aperture mask at near-field SPDC plane f2-signal"},
	{"single-slit-cntr-f2-sx", 4074, "FLOAT", 0,
	 "a single-slit aperture at near-field SPDC plane f2-signal in x-direction"},
	{"single-slit-cntr-f2-sy", 4075, "FLOAT", 0,
	 "a single-slit aperture at near-field SPDC plane f2-signal in y-direction"},
	{"single-slit-width-f2-sx", 4076, "FLOAT", 0,
	 "a single-slit aperture at near-field SPDC plane f2-signal in x-direction"},
	{"single-slit-width-f2-sy", 4077, "FLOAT", 0,
	 "a single-slit aperture at near-field SPDC plane f2-signal in y-direction"},
	{"double-slit-sep-f2-sx", 4078, "FLOAT", 0,
	 "a double-slit aperture at near-field SPDC plane f2-signal in x-direction"},
	{"double-slit-sep-f2-sy", 4079, "FLOAT", 0,
	 "a double-slit aperture at near-field SPDC plane f2-signal in y-direction"},
	{"double-slit-width-f2-sx", 4080, "FLOAT", 0,
	 "a double-slit aperture at near-field SPDC plane f2-signal in x-direction"},
	{"double-slit-width-f2-sy", 4081, "FLOAT", 0,
	 "a double-slit aperture at near-field SPDC plane f2-signal in y-direction"},
	{"knife-edge-pos-f2-ix", 4082, "FLOAT", 0,
	 "A knife edge aperture at near-field SPDC plane f2-idler in x-direction"},
	{"knife-edge-pos-f2-iy", 4083, "FLOAT", 0,
	 "A knife edge aperture at near-field SPDC plane f2-idler in y-direction"},
	{"knife-edge-invert-f2-ix", 4084, "INT", 0,
	 "A knife edge aperture at near-field SPDC plane f2-idler in x-direction"},
	{"knife-edge-invert-f2-iy", 4085, "INT", 0,
	 "A knife edge aperture at near-field SPDC plane f2-idler in y-direction"},
	{"circ-ap-r-f2-i", 4086, "FLOAT", 0,
	 "circular aperture mask at near-field SPDC plane f2-idler"},
	{"circ-ap-cntr-x-f2-i", 4087, "FLOAT", 0,
	 "circular aperture mask at near-field SPDC plane f2-idler"},
	{"circ-ap-cntr-y-f2-i", 4088, "FLOAT", 0,
	 "circular aperture mask at near-field SPDC plane f2-idler"},
	{"single-slit-cntr-f2-ix", 4089, "FLOAT", 0,
	 "a single-slit aperture at near-field SPDC plane f2-idler in x-direction"},
	{"single-slit-cntr-f2-iy", 4090, "FLOAT", 0,
	 "a single-slit aperture at near-field SPDC plane f2-idler in y-direction"},
	{"single-slit-width-f2-ix", 4091, "FLOAT", 0,
	 "a single-slit aperture at near-field SPDC plane f2-idler in x-direction"},
	{"single-slit-width-f2-iy", 4092, "FLOAT", 0,
	 "a single-slit aperture at near-field SPDC plane f2-idler in y-direction"},
	{"double-slit-sep-f2-ix", 4093, "FLOAT", 0,
	 "a double-slit aperture at near-field SPDC plane f2-idler in x-direction"},
	{"double-slit-sep-f2-iy", 4094, "FLOAT", 0,
	 "a double-slit aperture at near-field SPDC plane f2-idler in y-direction"},
	{"double-slit-width-f2-ix", 4095, "FLOAT", 0,
	 "a double-slit aperture at near-field SPDC plane f2-idler in x-direction"},
	{"double-slit-width-f2-iy", 4096, "FLOAT", 0,
	 "a double-slit aperture at near-field SPDC plane f2-idler in y-direction"},
	{"pump-pump-mask-type", 4097, "INT", 0, "Mask type at plane pump-pump"},
	{"f1-signal-mask-type", 4098, "INT", 0, "Mask type at plane f1-signal"},
	{"f1-idler-mask-type", 4099, "INT", 0, "Mask type at plane f1-idler"},
	{"f2-signal-mask-type", 4100, "INT", 0, "Mask type at plane f2-signal"},
	{"f2-idler-mask-type", 4101, "INT", 0, "Mask type at plane f2-idler"},
	{0}
};

static error_t parse_opt(int key, char *arg, struct argp_state *state)
{
	struct spdc_prm *prm = state->input;
	if (key == 4000)
		prm->lambda_pump = atof(arg);
	else if (key == 4001)
		prm->lambda_signal = atof(arg);
	else if (key == 4002)
		prm->lambda_idler = atof(arg);
	else if (key == 4003)
		prm->hermite_gauss_m = atoi(arg);
	else if (key == 4004)
		prm->hermite_gauss_n = atoi(arg);
	else if (key == 4005)
		prm->pump_w0 = atof(arg);
	else if (key == 4006)
		prm->pump_z = atof(arg);
	else if (key == 4007)
		prm->pseudo_tem = atoi(arg);
	else if (key == 4008)
		prm->width = atof(arg);
	else if (key == 4009)
		prm->length = atof(arg);
	else if (key == 4010)
		prm->theta = atof(arg);
	else if (key == 4011)
		prm->phase = atoi(arg);
	else if (key == 4012)
		prm->phase_compensation = atoi(arg);
	else if (key == 4013)
		prm->phase_match_type = atoi(arg);
	else if (key == 4014)
		prm->N = atoi(arg);
	else if (key == 4015)
		prm->schmidt_rank = atoi(arg);
	else if (key == 4016)
		prm->eigenvalues_only = atoi(arg);
	else if (key == 4017)
		prm->focal_length = atof(arg);
	else if (key == 4018)
		prm->nf_signal_offset = atof(arg);
	else if (key == 4019)
		prm->nf_idler_offset = atof(arg);
	else if (key == 4020)
		prm->ff_signal_offset = atof(arg);
	else if (key == 4021)
		prm->ff_idler_offset = atof(arg);
	else if (key == 4022)
		prm->knife_edge_pos_pump_px = atof(arg);
	else if (key == 4023)
		prm->knife_edge_pos_pump_py = atof(arg);
	else if (key == 4024)
		prm->knife_edge_invert_pump_px = atoi(arg);
	else if (key == 4025)
		prm->knife_edge_invert_pump_py = atoi(arg);
	else if (key == 4026)
		prm->circ_ap_r_pump_p = atof(arg);
	else if (key == 4027)
		prm->circ_ap_cntr_x_pump_p = atof(arg);
	else if (key == 4028)
		prm->circ_ap_cntr_y_pump_p = atof(arg);
	else if (key == 4029)
		prm->single_slit_cntr_pump_px = atof(arg);
	else if (key == 4030)
		prm->single_slit_cntr_pump_py = atof(arg);
	else if (key == 4031)
		prm->single_slit_width_pump_px = atof(arg);
	else if (key == 4032)
		prm->single_slit_width_pump_py = atof(arg);
	else if (key == 4033)
		prm->double_slit_sep_pump_px = atof(arg);
	else if (key == 4034)
		prm->double_slit_sep_pump_py = atof(arg);
	else if (key == 4035)
		prm->double_slit_width_pump_px = atof(arg);
	else if (key == 4036)
		prm->double_slit_width_pump_py = atof(arg);
	else if (key == 4037)
		prm->knife_edge_pos_f1_sx = atof(arg);
	else if (key == 4038)
		prm->knife_edge_pos_f1_sy = atof(arg);
	else if (key == 4039)
		prm->knife_edge_invert_f1_sx = atoi(arg);
	else if (key == 4040)
		prm->knife_edge_invert_f1_sy = atoi(arg);
	else if (key == 4041)
		prm->circ_ap_r_f1_s = atof(arg);
	else if (key == 4042)
		prm->circ_ap_cntr_x_f1_s = atof(arg);
	else if (key == 4043)
		prm->circ_ap_cntr_y_f1_s = atof(arg);
	else if (key == 4044)
		prm->single_slit_cntr_f1_sx = atof(arg);
	else if (key == 4045)
		prm->single_slit_cntr_f1_sy = atof(arg);
	else if (key == 4046)
		prm->single_slit_width_f1_sx = atof(arg);
	else if (key == 4047)
		prm->single_slit_width_f1_sy = atof(arg);
	else if (key == 4048)
		prm->double_slit_sep_f1_sx = atof(arg);
	else if (key == 4049)
		prm->double_slit_sep_f1_sy = atof(arg);
	else if (key == 4050)
		prm->double_slit_width_f1_sx = atof(arg);
	else if (key == 4051)
		prm->double_slit_width_f1_sy = atof(arg);
	else if (key == 4052)
		prm->knife_edge_pos_f1_ix = atof(arg);
	else if (key == 4053)
		prm->knife_edge_pos_f1_iy = atof(arg);
	else if (key == 4054)
		prm->knife_edge_invert_f1_ix = atoi(arg);
	else if (key == 4055)
		prm->knife_edge_invert_f1_iy = atoi(arg);
	else if (key == 4056)
		prm->circ_ap_r_f1_i = atof(arg);
	else if (key == 4057)
		prm->circ_ap_cntr_x_f1_i = atof(arg);
	else if (key == 4058)
		prm->circ_ap_cntr_y_f1_i = atof(arg);
	else if (key == 4059)
		prm->single_slit_cntr_f1_ix = atof(arg);
	else if (key == 4060)
		prm->single_slit_cntr_f1_iy = atof(arg);
	else if (key == 4061)
		prm->single_slit_width_f1_ix = atof(arg);
	else if (key == 4062)
		prm->single_slit_width_f1_iy = atof(arg);
	else if (key == 4063)
		prm->double_slit_sep_f1_ix = atof(arg);
	else if (key == 4064)
		prm->double_slit_sep_f1_iy = atof(arg);
	else if (key == 4065)
		prm->double_slit_width_f1_ix = atof(arg);
	else if (key == 4066)
		prm->double_slit_width_f1_iy = atof(arg);
	else if (key == 4067)
		prm->knife_edge_pos_f2_sx = atof(arg);
	else if (key == 4068)
		prm->knife_edge_pos_f2_sy = atof(arg);
	else if (key == 4069)
		prm->knife_edge_invert_f2_sx = atoi(arg);
	else if (key == 4070)
		prm->knife_edge_invert_f2_sy = atoi(arg);
	else if (key == 4071)
		prm->circ_ap_r_f2_s = atof(arg);
	else if (key == 4072)
		prm->circ_ap_cntr_x_f2_s = atof(arg);
	else if (key == 4073)
		prm->circ_ap_cntr_y_f2_s = atof(arg);
	else if (key == 4074)
		prm->single_slit_cntr_f2_sx = atof(arg);
	else if (key == 4075)
		prm->single_slit_cntr_f2_sy = atof(arg);
	else if (key == 4076)
		prm->single_slit_width_f2_sx = atof(arg);
	else if (key == 4077)
		prm->single_slit_width_f2_sy = atof(arg);
	else if (key == 4078)
		prm->double_slit_sep_f2_sx = atof(arg);
	else if (key == 4079)
		prm->double_slit_sep_f2_sy = atof(arg);
	else if (key == 4080)
		prm->double_slit_width_f2_sx = atof(arg);
	else if (key == 4081)
		prm->double_slit_width_f2_sy = atof(arg);
	else if (key == 4082)
		prm->knife_edge_pos_f2_ix = atof(arg);
	else if (key == 4083)
		prm->knife_edge_pos_f2_iy = atof(arg);
	else if (key == 4084)
		prm->knife_edge_invert_f2_ix = atoi(arg);
	else if (key == 4085)
		prm->knife_edge_invert_f2_iy = atoi(arg);
	else if (key == 4086)
		prm->circ_ap_r_f2_i = atof(arg);
	else if (key == 4087)
		prm->circ_ap_cntr_x_f2_i = atof(arg);
	else if (key == 4088)
		prm->circ_ap_cntr_y_f2_i = atof(arg);
	else if (key == 4089)
		prm->single_slit_cntr_f2_ix = atof(arg);
	else if (key == 4090)
		prm->single_slit_cntr_f2_iy = atof(arg);
	else if (key == 4091)
		prm->single_slit_width_f2_ix = atof(arg);
	else if (key == 4092)
		prm->single_slit_width_f2_iy = atof(arg);
	else if (key == 4093)
		prm->double_slit_sep_f2_ix = atof(arg);
	else if (key == 4094)
		prm->double_slit_sep_f2_iy = atof(arg);
	else if (key == 4095)
		prm->double_slit_width_f2_ix = atof(arg);
	else if (key == 4096)
		prm->double_slit_width_f2_iy = atof(arg);
	else if (key == 4097)
		prm->pump_pump_mask_type = atoi(arg);
	else if (key == 4098)
		prm->f1_signal_mask_type = atoi(arg);
	else if (key == 4099)
		prm->f1_idler_mask_type = atoi(arg);
	else if (key == 4100)
		prm->f2_signal_mask_type = atoi(arg);
	else if (key == 4101)
		prm->f2_idler_mask_type = atoi(arg);
	else
		return (ARGP_ERR_UNKNOWN);

	return (0);
}

/*[[[end]]]*/

static struct argp argp = { options, parse_opt, args_doc, doc };

int main(int argc, char **argv)
{
	/* Default spdc parameters and argument parsing */
	struct spdc_prm prm;
	spdc_parse_args(&argp, argc, argv, &prm);
	spdc_dump_prm(prm);

	/* Initialize auxilary parameters */
	double rho_norm = pow(prm.width / prm.N, 4);
	double q_norm = pow(1.0 / prm.width, 4);

	/* FFTW multithreading */
	fftw_init_threads();
	fftw_plan_with_nthreads(omp_get_max_threads());

	/* Initialize arrays */
	struct spdc_mdarray *A[12];
	spdc_prepare_working_arrays(A, prm);

	/* Initialize SPDC field coefficients */
	field_co flds[3];
	spdc_calculate_field_parameters(flds, prm);

	/* Calculate initial field distribution */
	spdc_prepare_input_fields(A[1], A[3], prm);

	/* Compute angular spectrum of pump beam */
	spdc_fftw(A[3], FFTW_FORWARD, sqrt(rho_norm));

	/* Compute q-vector space */
	spdc_compute_q_space(A[2], prm);

	/* Calculate Phi */
	spdc_phi(A[0], A[3], A[2], flds, prm);

	/* Apply spectrum mask */
	spdc_mask(A[2], A[2], A[2], A[2], A[0], 1, &prm);

	/* Perform Schmidt decomposition. Note that the program terminates
	 * after calculating the Schmidt decomposition. Any propagation can be
	 * more conveniently performed by manipulating the Schmidt modes. */
	if (prm.schmidt_rank > 0) {
		spdc_schmidt_decomposition(A[0], (size_t) prm.schmidt_rank,
					   prm);
		return (0);
	}

	/* Calculate reduced coincidence count rate for signal */
	int sum_axes[2] = { 2, 3 };
	spdc_sum_reduce(A[0], 2, sum_axes, A[4], sqrt(q_norm), 1);

	/* Calculate reduced coincidence count rate for idler */
	sum_axes[0] = 0;
	sum_axes[1] = 1;
	spdc_sum_reduce(A[0], 2, sum_axes, A[5], sqrt(q_norm), 1);

	/* TODO: Apply near field detector plane offset */
	spdc_near_field_offset(A[0], A[2], flds, prm);

	/* Transform to spatial domain */
	spdc_fftw(A[0], FFTW_BACKWARD, q_norm);

	/* Calculate idler near field */
	spdc_sum_reduce(A[0], 2, sum_axes, A[7], sqrt(rho_norm), 1);

	/* Calculate signal near field */
	sum_axes[0] = 2;
	sum_axes[1] = 3;
	spdc_sum_reduce(A[0], 2, sum_axes, A[6], sqrt(rho_norm), 1);

	/* Apply mask in real space */
	spdc_mask(A[1], A[1], A[1], A[1], A[0], 2, &prm);

	/* Calculate signal coincidence rate */
	spdc_sum_reduce(A[0], 2, sum_axes, A[8], sqrt(rho_norm), 1);

	/* Calculate signal near field */
	sum_axes[0] = 0;
	sum_axes[1] = 1;
	spdc_sum_reduce(A[0], 2, sum_axes, A[9], sqrt(rho_norm), 1);

	/* TODO: Apply far-field detector plane offset */

	/* Transform back to far-field pattern to obtain interference fringes */
	spdc_fftw(A[0], FFTW_FORWARD, rho_norm);

	/* Idler far-field pattern with the spatial mask applied */
	spdc_sum_reduce(A[0], 2, sum_axes, A[11], sqrt(q_norm), 1);

	/* Signal far-field pattern with the spatial mask applied */
	sum_axes[0] = 2;
	sum_axes[1] = 3;
	spdc_sum_reduce(A[0], 2, sum_axes, A[10], sqrt(q_norm), 1);

	/* Dump all arrays to stdout */
	spdc_dump_all_arrays(A + 1, 11);

	return (0);
}

int spdc_parse_args(struct argp *argp, int argc, char **argv,
		    struct spdc_prm *prm)
{
	/* Default spdc parameters */
	/*[[[cog
	   import cog_codegen

	   options = cog_codegen.get_options()
	   cog_codegen.set_defaults(options)
	   ]]] */
	prm->lambda_pump = 4.05e-07;
	prm->lambda_signal = 8.1e-07;
	prm->lambda_idler = 8.1e-07;
	prm->hermite_gauss_m = 1;
	prm->hermite_gauss_n = 0;
	prm->pump_w0 = 7e-05;
	prm->pump_z = 0.0;
	prm->pseudo_tem = 0;
	prm->width = 0.0008;
	prm->length = 0.002;
	prm->theta = 42.4;
	prm->phase = 1;
	prm->phase_compensation = 0;
	prm->phase_match_type = 2;
	prm->N = 64;
	prm->schmidt_rank = 0;
	prm->eigenvalues_only = 0;
	prm->focal_length = 50;
	prm->nf_signal_offset = 0.0;
	prm->nf_idler_offset = 0.0;
	prm->ff_signal_offset = 0.0;
	prm->ff_idler_offset = 0.0;
	prm->knife_edge_pos_pump_px = 0;
	prm->knife_edge_pos_pump_py = 0;
	prm->knife_edge_invert_pump_px = 0;
	prm->knife_edge_invert_pump_py = 0;
	prm->circ_ap_r_pump_p = 1.4e-05;
	prm->circ_ap_cntr_x_pump_p = 0;
	prm->circ_ap_cntr_y_pump_p = 0;
	prm->single_slit_cntr_pump_px = 0;
	prm->single_slit_cntr_pump_py = 0;
	prm->single_slit_width_pump_px = 0;
	prm->single_slit_width_pump_py = 0;
	prm->double_slit_sep_pump_px = 0.000105;
	prm->double_slit_sep_pump_py = 0.000105;
	prm->double_slit_width_pump_px = 3e-05;
	prm->double_slit_width_pump_py = 3e-05;
	prm->knife_edge_pos_f1_sx = 0;
	prm->knife_edge_pos_f1_sy = 0;
	prm->knife_edge_invert_f1_sx = 0;
	prm->knife_edge_invert_f1_sy = 0;
	prm->circ_ap_r_f1_s = 1.4e-05;
	prm->circ_ap_cntr_x_f1_s = 0;
	prm->circ_ap_cntr_y_f1_s = 0;
	prm->single_slit_cntr_f1_sx = 0;
	prm->single_slit_cntr_f1_sy = 0;
	prm->single_slit_width_f1_sx = 0;
	prm->single_slit_width_f1_sy = 0;
	prm->double_slit_sep_f1_sx = 0.000105;
	prm->double_slit_sep_f1_sy = 0.000105;
	prm->double_slit_width_f1_sx = 3e-05;
	prm->double_slit_width_f1_sy = 3e-05;
	prm->knife_edge_pos_f1_ix = 0;
	prm->knife_edge_pos_f1_iy = 0;
	prm->knife_edge_invert_f1_ix = 0;
	prm->knife_edge_invert_f1_iy = 0;
	prm->circ_ap_r_f1_i = 1.4e-05;
	prm->circ_ap_cntr_x_f1_i = 0;
	prm->circ_ap_cntr_y_f1_i = 0;
	prm->single_slit_cntr_f1_ix = 0;
	prm->single_slit_cntr_f1_iy = 0;
	prm->single_slit_width_f1_ix = 0;
	prm->single_slit_width_f1_iy = 0;
	prm->double_slit_sep_f1_ix = 0.000105;
	prm->double_slit_sep_f1_iy = 0.000105;
	prm->double_slit_width_f1_ix = 3e-05;
	prm->double_slit_width_f1_iy = 3e-05;
	prm->knife_edge_pos_f2_sx = 0;
	prm->knife_edge_pos_f2_sy = 0;
	prm->knife_edge_invert_f2_sx = 0;
	prm->knife_edge_invert_f2_sy = 0;
	prm->circ_ap_r_f2_s = 1.4e-05;
	prm->circ_ap_cntr_x_f2_s = 0;
	prm->circ_ap_cntr_y_f2_s = 0;
	prm->single_slit_cntr_f2_sx = 0;
	prm->single_slit_cntr_f2_sy = 0;
	prm->single_slit_width_f2_sx = 0;
	prm->single_slit_width_f2_sy = 0;
	prm->double_slit_sep_f2_sx = 0.000105;
	prm->double_slit_sep_f2_sy = 0.000105;
	prm->double_slit_width_f2_sx = 3e-05;
	prm->double_slit_width_f2_sy = 3e-05;
	prm->knife_edge_pos_f2_ix = 0;
	prm->knife_edge_pos_f2_iy = 0;
	prm->knife_edge_invert_f2_ix = 0;
	prm->knife_edge_invert_f2_iy = 0;
	prm->circ_ap_r_f2_i = 1.4e-05;
	prm->circ_ap_cntr_x_f2_i = 0;
	prm->circ_ap_cntr_y_f2_i = 0;
	prm->single_slit_cntr_f2_ix = 0;
	prm->single_slit_cntr_f2_iy = 0;
	prm->single_slit_width_f2_ix = 0;
	prm->single_slit_width_f2_iy = 0;
	prm->double_slit_sep_f2_ix = 0.000105;
	prm->double_slit_sep_f2_iy = 0.000105;
	prm->double_slit_width_f2_ix = 3e-05;
	prm->double_slit_width_f2_iy = 3e-05;
	prm->pump_pump_mask_type = 0;
	prm->f1_signal_mask_type = 0;
	prm->f1_idler_mask_type = 0;
	prm->f2_signal_mask_type = 0;
	prm->f2_idler_mask_type = 0;
	/*[[[end]]] */

	/* Argument parsing */
	argp_parse(argp, argc, argv, 0, 0, prm);

	/*TODO: Sanity checking and convenience conversions */
	prm->theta = prm->theta / 180.0 * M_PI;
	return (SPDC_RETURN_OK);
}

/* Apply the necessary phase shift to the far-field biphoton amplitude to
 * effect a longitudinal displacement of the signal and idler planes.
 *
 * :param A:    The complex 4D biphoton amplitude
 * :param q:    The transverse momentum coordinate in the far-field
 * :param flds: An array[3] of field structures containing the coefficients
 *              for the pump, signal and idler fields in this order.
 * :param prm:  The parameter structure containing the offsets and the focal
 *              length of the lenses.
 *
 */
int spdc_near_field_offset(spdc_mdarray * A, spdc_mdarray * q, field_co * flds,
			   struct spdc_prm prm)
{
	size_t N = prm.N;
	size_t n_elements = spdc_get_size(A);
	double d_s = prm.nf_signal_offset;
	double d_i = prm.nf_idler_offset;

#pragma omp parallel
	{
		double q_sx, q_sy, q_ix, q_iy;
		complex double cur_item;
		ptrdiff_t multi_idx[4];

#pragma omp for
		for (size_t i = 0; i < n_elements; i++) {
			spdc_unravel(A, i, multi_idx);
			cur_item = SPDC_GET_COMPLEX(*A, i);
			q_sx = SPDC_GET_DOUBLE(*q, multi_idx[0]);
			q_sy = SPDC_GET_DOUBLE(*q, multi_idx[1]);
			q_ix = SPDC_GET_DOUBLE(*q, multi_idx[2]);
			q_iy = SPDC_GET_DOUBLE(*q, multi_idx[3]);
			cur_item *=
			    cexp(-I * d_s * (q_sx * q_sx + q_sy * q_sy) /
				 (2.0 * flds[1].k));
			cur_item *=
			    cexp(-I * d_i * (q_ix * q_ix + q_iy * q_iy) /
				 (2.0 * flds[2].k));
			SPDC_GET_COMPLEX(*A, i) = cur_item;
		}
	}

	return SPDC_RETURN_OK;
}

/* Prepare the necessary storage for storing results and buffer calculations.
 * The arrays are kept in a array of pointers with the following conventions:
 *
 *      * A[0] - 4D workbench
 *      * A[1] - Transverse spatial coordinate rho
 *      * A[2] - Transverse momentum coordinate q
 *      * A[3] - Input field (2D)
 *      * A[4] - A[11] - Reduced coincidence rates C^(2) for signal and idler
 *        at the different position through the setup
 *
 * */
int spdc_prepare_working_arrays(struct spdc_mdarray *arrays[12],
				struct spdc_prm prm)
{
	size_t shape_1d[1] = { prm.N };
	size_t shape_2d[2] = { prm.N, prm.N };
	size_t shape_4d[4] = { prm.N, prm.N, prm.N, prm.N };

	for (int i = 0; i < 12; i++) {
		arrays[i] =
		    (struct spdc_mdarray *)malloc(sizeof(struct spdc_mdarray));
	}

	spdc_init_mdarray(4, shape_4d, SPDC_DTYPE_COMPLEX, arrays[0]);

	for (int i = 0; i < 2; i++) {
		spdc_init_mdarray(1, shape_1d, SPDC_DTYPE_DOUBLE,
				  arrays[i + 1]);
	}

	for (int i = 0; i < 9; i++) {
		spdc_init_mdarray(2, shape_2d, SPDC_DTYPE_COMPLEX,
				  arrays[i + 3]);
	}
	return (SPDC_RETURN_OK);
}

/* Caluclate refractive indices and walkoff coefficients for the fields */
int spdc_calculate_field_parameters(struct field_co *flds, struct spdc_prm prm)
{
	double n_o, n_e;
	const double lambda[3] = { prm.lambda_pump, prm.lambda_signal,
		prm.lambda_idler
	};

	for (int i = 0; i < 3; i++) {
		spdc_sellmeier_bbo(lambda[i], &n_o, &n_e);
		flds[i].n_o = n_o;
		flds[i].n_e = n_e;
		spdc_walkoff_coefficients(n_o, n_e, prm.theta, (flds + i));
		flds[i].omega = 2 * M_PI * SPDC_C0 / lambda[i];

		flds[i].lambda = lambda[i];
		flds[i].k = 2 * M_PI / flds[i].lambda;
	}
	return (SPDC_RETURN_OK);
}

/* Dump arrays to stdout, separated by "end of array" (eoa) lines*/
int spdc_dump_all_arrays(spdc_mdarray ** arrays, int n_arrays)
{
	for (int i = 0; i < n_arrays; i++) {
		spdc_dump_array(arrays[i]);
		if (i < (n_arrays - 1))
			printf("eoa\n");
	}
	return (SPDC_RETURN_OK);
}

/* Compute initial field distribution and apply an eventual knife-edge mask.
 * The fields are in native fft order. */
int spdc_prepare_input_fields(spdc_mdarray * rho, spdc_mdarray * u0, const struct
			      spdc_prm prm)
{
	size_t n_items = spdc_get_size(u0);
	int retcode = 0;

	spdc_linspace(-prm.width / 2.0, prm.width / 2.0, rho);
	spdc_fftshift(rho);
	if (prm.pseudo_tem == 0) {
		retcode =
		    spdc_gauss_hermite(prm.hermite_gauss_m, prm.hermite_gauss_n,
				       rho, rho, prm.pump_z, prm.pump_w0,
				       prm.lambda_pump, u0);
	} else if (prm.pseudo_tem == 1) {
		retcode =
		    spdc_pseudo_tem10(rho, rho, prm.pump_w0, prm.lambda_pump,
				      u0);
	}

	if (retcode != SPDC_RETURN_OK)
		return (retcode);

	spdc_mask_2d(rho, rho, u0, &prm);
	return (SPDC_RETURN_OK);
}

int spdc_compute_q_space(spdc_mdarray * q, struct spdc_prm prm)
{
	size_t n_items = spdc_get_size(q);
	spdc_fftfreq(q, prm.width);
	for (ptrdiff_t i = 0; i < n_items; i++) {
		SPDC_GET_DOUBLE(*q, i) *= 2 * M_PI;
	}
	return (SPDC_RETURN_OK);
}

/* Compute the forward DFT of arr in place, the parameter direction can be
 * either FFTW_FORWARD or FFTW_BACKWARD. */
int spdc_fftw(spdc_mdarray * arr, int direction, double complex normalization)
{
	assert(arr->dtype == SPDC_DTYPE_COMPLEX);

	int shape[arr->ndim];
	size_t n = spdc_get_size(arr);
	fftw_plan p;

	for (int i = 0; i < arr->ndim; i++) {
		shape[i] = (int)arr->shape[i];
	}
	p = fftw_plan_dft(arr->ndim, shape, arr->data, arr->data, direction,
			  FFTW_ESTIMATE);
	fftw_execute(p);

#pragma omp parallel for
	for (ptrdiff_t i = 0; i < n; i++) {
		SPDC_GET_COMPLEX(*arr, i) *= normalization;
	}

	return (SPDC_RETURN_OK);

}

/* Returns the pump spectrum v[i,j,k,l] = v(q_s[i,j] + q_i[k,l]) where
 * v(q) is the angular spectrum of the pump beam in native FFT order. v(q) is
 * assumed to be zero everywhere except -Q_max < q_s + q_i < Q_max where Q_max
 * is the largest wave vector accessible by the DFT (which usually means 1 /
 * dx.*/
inline double complex spdc_pump_spectrum(spdc_mdarray * v, struct spdc_prm prm,
					 ptrdiff_t i, ptrdiff_t j, ptrdiff_t k,
					 ptrdiff_t l)
{
	int zero_flag = 0;
	ptrdiff_t multi_idx[2];
	double complex res = 0;

	/* x-component */
	if (i < prm.N / 2 && k < prm.N / 2) {
		if (i + k < prm.N / 2)
			multi_idx[0] = i + k;
		else
			zero_flag = 1;
	}

	if (i >= prm.N / 2 && k >= prm.N / 2) {
		if (i + k >= 3 * prm.N / 2)
			multi_idx[0] = i + k - prm.N;
		else
			zero_flag = 1;
	}

	if (i + k < prm.N)
		multi_idx[0] = i + k;
	else if (i + k >= prm.N)
		multi_idx[0] = i + k - prm.N;
	else
		assert(0);

	/* y-component */
	if (j < prm.N / 2 && l < prm.N / 2) {
		if (j + l < prm.N / 2)
			multi_idx[1] = j + l;
		else
			zero_flag = 1;
	}

	if (j >= prm.N / 2 && l >= prm.N / 2) {
		if (j + l >= 3 * prm.N / 2)
			multi_idx[1] = j + l - prm.N;
		else
			zero_flag = 1;
	}

	if (j + l < prm.N)
		multi_idx[1] = j + l;
	else if (j + l >= prm.N)
		multi_idx[1] = j + l - prm.N;
	else
		assert(0);

	if (zero_flag)
		res = 0.0;
	else {
		res = spdc_get_complex(v, multi_idx);
	}
	return (res);

}

/* Compute the transverse wave vector spectrum of the SPDC light phi(q_s, q_i) */
int spdc_phi(spdc_mdarray * phi, spdc_mdarray * v, spdc_mdarray * q,
	     struct field_co *flds, struct spdc_prm prm)
{
	size_t n = spdc_get_size(phi);
	double C_p, C_s, C_i;
	const double *q_ = (double *)q->data;

	/* Calculate coefficients for exp phase term compensation */
	if (prm.phase_compensation == 1 && prm.phase == 1) {
		C_p =
		    prm.length * SPDC_C0 / (flds[0].eta * flds[0].omega * 4.0);
		C_s =
		    prm.length * SPDC_C0 / (flds[1].n_o * flds[1].omega * 4.0);
		C_i =
		    prm.length * SPDC_C0 / (flds[2].eta * flds[2].omega * 4.0);
	}
#pragma omp parallel
	{
		ptrdiff_t i, j, k, l;
		double complex tmp, delta_kz;
		double tmp_phase;

#pragma omp for
		for (ptrdiff_t m = 0; m < n; m++) {
			SPDC_UNRAVEL4(*phi, m, i, j, k, l);

			/* Calculate longitudinal phase mismatch */
			if (prm.phase_match_type == 2) {
				delta_kz =
				    spdc_compute_delta_kz_type2(q, i, j, k, l,
								flds);
			} else if (prm.phase_match_type == 1) {
				delta_kz =
				    spdc_compute_delta_kz_type1(q, i, j, k, l,
								flds);
			} else if (prm.phase_match_type == 0) {
                delta_kz = spdc_compute_delta_kz_type0(q, i, j, k, l, flds);
                delta_kz += prm.theta;
            }


			/* Phase matching condition */
			tmp = delta_kz * prm.length / 2.0;
			if (cabs(tmp) != 0.0)
				tmp = csin(tmp) / tmp;
			else
				tmp = 1.0;

			/* Apply exponential phase after discussion with John */
			if (prm.phase == 1) {
				tmp *= cexp(-I * delta_kz * prm.length / 2.0);
			}

			/* Apply exponential phase term compensation by longitudinal
			 * and transversal shifting of the near-field planes.  */
			if (prm.phase_compensation == 1 && prm.phase == 1) {
				tmp_phase =
				    C_s * (q_[i] * q_[i] + q_[j] * q_[j]);
				tmp_phase +=
				    C_i * (q_[k] * q_[k] + q_[l] * q_[l]);
				tmp_phase +=
				    flds[0].alpha * prm.length / 2.0 * (q_[i] +
									q_[k]);
				tmp_phase -=
				    flds[2].alpha * prm.length / 2.0 * q_[k];
				tmp *= cexp(I * tmp_phase);
			}

			/* Angular spectrum from pump beam */
			tmp *= spdc_pump_spectrum(v, prm, i, j, k, l);
			SPDC_GET_COMPLEX(*phi, m) = tmp;
		}
	}
	return (SPDC_RETURN_OK);
}

/* Calculate the longitudinal wave vector mismatch for type-II SPDC dk_z =
 * dk_z(q_s[i,j], q_i[k,l]) with q_s = q_i = q */
inline double complex spdc_compute_delta_kz_type2(spdc_mdarray * q, ptrdiff_t i,
						  ptrdiff_t j, ptrdiff_t k,
						  ptrdiff_t l,
						  struct field_co *flds)
{
	double *ptr_d = (double *)q->data;
	double complex res;
	res =
	    flds[0].eta * flds[0].omega / SPDC_C0 + flds[0].alpha * (ptr_d[i] +
								     ptr_d[k])
	    -
	    (SPDC_C0 / (2.0 * flds[0].eta * flds[0].omega)) *
	    (pow(ptr_d[i] + ptr_d[k], 2) + pow(ptr_d[j] + ptr_d[l], 2));

	res -=
	    flds[2].eta * flds[2].omega / SPDC_C0 + flds[2].alpha * ptr_d[k] -
	    (SPDC_C0 / (2.0 * flds[2].eta * flds[2].omega)) *
	    (pow(ptr_d[k], 2) + pow(ptr_d[l], 2));

	res -=
	    flds[1].n_o * flds[1].omega / SPDC_C0 -
	    (SPDC_C0 / (2.0 * flds[1].n_o * flds[1].omega)) *
	    (pow(ptr_d[i], 2) + pow(ptr_d[j], 2));

	return (res);
}

/* Calculate the longitudinal wave vector mismatch for type-I SPDC dk_z =
 * dk_z(q_s[i,j], q_i[k,l]) with q_s = q_i = q */
inline double complex spdc_compute_delta_kz_type1(spdc_mdarray * q, ptrdiff_t i,
						  ptrdiff_t j, ptrdiff_t k,
						  ptrdiff_t l,
						  struct field_co *flds)
{
	double *ptr_d = (double *)q->data;
	double complex res;
	res =
	    flds[0].eta * flds[0].omega / SPDC_C0 + flds[0].alpha * (ptr_d[i] +
								     ptr_d[k])
	    -
	    (SPDC_C0 / (2.0 * flds[0].eta * flds[0].omega)) *
	    (pow(ptr_d[i] + ptr_d[k], 2) + pow(ptr_d[j] + ptr_d[l], 2));

	res -=
	    flds[2].n_o * flds[2].omega / SPDC_C0 -
	    (SPDC_C0 / (2.0 * flds[2].n_o * flds[2].omega)) *
	    (pow(ptr_d[k], 2) + pow(ptr_d[l], 2));

	res -=
	    flds[1].n_o * flds[1].omega / SPDC_C0 -
	    (SPDC_C0 / (2.0 * flds[1].n_o * flds[1].omega)) *
	    (pow(ptr_d[i], 2) + pow(ptr_d[j], 2));

	return (res);
}


/* Calculate the longitudinal wave vector mismatch for a virtual phase
 * matching condition with no crystal anisotropy and a freely tunable phase
 * mismatch theta (the OA angle theta is then rededicated to this phase
 * mismatch). 
 * dk_z(q_s[i,j], q_i[k,l]) with q_s = q_i = q */
inline double complex spdc_compute_delta_kz_type0(spdc_mdarray *q, ptrdiff_t
        i, ptrdiff_t j, ptrdiff_t k, ptrdiff_t l, struct field_co *flds)
{
	double *ptr_d = (double *)q->data;
	double complex res;
    res = - (SPDC_C0 / (2.0 * flds[0].omega)) * (pow(ptr_d[i] + ptr_d[k], 2) +
            pow(ptr_d[j] + ptr_d[l], 2));

    res -= - (SPDC_C0 / (2.0 * flds[2].omega)) * (pow(ptr_d[k], 2) +
            pow(ptr_d[l], 2));

    res -= - (SPDC_C0 / (2.0 * flds[1].omega)) * (pow(ptr_d[i], 2) +
            pow(ptr_d[j], 2));

	return (res);
}

/*
 * Calculate the Schmidt decomposition of the 4D biphoton amplitude phi.
 * */
int spdc_schmidt_decomposition(spdc_mdarray * phi, size_t k,
			       struct spdc_prm prm)
{
	/* Sanity checking */
	assert(phi->dtype == SPDC_DTYPE_COMPLEX);

	size_t m = prm.N * prm.N;
	size_t n = m;
	size_t l = k + 2;
	size_t n_iter = 2;
	double E0;
	double Ek;

	gsl_matrix_complex *U;
	gsl_matrix_complex *Va;
	gsl_vector *S;

	/* Create 2D matrix view from 4D biphoton amplitude. The cast to double
	 * instead of complex is on purpose as the bookeeping happens inside the
	 * view_array function. */
	double *phi_pointer = (double *)phi->data;
	gsl_matrix_complex_view phi_view =
	    gsl_matrix_complex_view_array(phi_pointer, m, n);
	ssvd_ssvd(&(phi_view.matrix), k, l, n_iter, &U, &Va, &S);

	/* Normalize eigenvalues */
	gsl_vector_complex_view v_view =
	    gsl_vector_complex_view_array(phi_pointer, m * n);
	E0 = gsl_blas_dznrm2(&(v_view.vector));
	gsl_blas_dscal(1.0 / E0, S);
	Ek = gsl_blas_dnrm2(S);
	/* printf("E0: %e \n", E0 * E0);
	   printf("Ek: %e \n", Ek * Ek);
	 */

	/* Dumping matrices to stdout */
	if (prm.eigenvalues_only == 0) {
		gsl_matrix_complex_fprintf(stdout, U, "%.5e");
		printf("EOA\n");
		gsl_matrix_complex_fprintf(stdout, Va, "%.5e");
		printf("EOA\n");
	}
	gsl_vector_fprintf(stdout, S, "%.5e");
	printf("EOA\n");
}
